#include "Decoder.h"
#include "Config.h"
#include "Utils.h"
#include <limits>


Word* Path::p_StartWord = NULL;
int   Path::i_InitialStartWords = 0;

//-----------------------------------------------------------
Path::Path (void)
{
    d_Score = -1000000000;
    p_FeatureVector = NULL;
    b_LastWordIsNotInSection = false;
    b_YGoodMatch =  true;
    
    assert (i_InitialStartWords > 0);
    for (int i = 0; i < i_InitialStartWords; ++ i)
		dq_Words.push_back (p_StartWord);

    memset (arr_SectionFeatureTotals, 0, FEATURE_COUNT * sizeof (double));
    memset (arr_SectionFeatureCounts, 0, FEATURE_COUNT * sizeof (long));
}


//-----------------------------------------------------------
Path::Path (Path& _rOld)
{
	d_Score = _rOld.d_Score;
	if (NULL != _rOld.p_FeatureVector)
	    p_FeatureVector = new FeatureVector (*_rOld.p_FeatureVector);
	else
	    p_FeatureVector = NULL;
	dq_Words = _rOld.dq_Words;
	b_LastWordIsNotInSection = _rOld.b_LastWordIsNotInSection;
	b_YGoodMatch = _rOld.b_YGoodMatch;
	hmp_TitleStems = _rOld.hmp_TitleStems;

	for (int i = 0; i < FEATURE_COUNT; ++ i)
	{
	    arr_SectionFeatureTotals [i] = _rOld.arr_SectionFeatureTotals [i];
	    arr_SectionFeatureCounts [i] = _rOld.arr_SectionFeatureCounts [i];
	}
}


//-----------------------------------------------------------
Path::~Path (void)
{
	if (NULL != p_FeatureVector)
	    delete p_FeatureVector;
	dq_Words.clear ();
}


//-----------------------------------------------------------
void Path::Grow (Word* _pWord, double _dScore, FeatureVector* _pFeatureVector, bool _bWordIsNotInSection)
{
	d_Score = _dScore;
	if (p_FeatureVector != _pFeatureVector)
	{
	    if (NULL != p_FeatureVector)
		delete p_FeatureVector;
	    p_FeatureVector = _pFeatureVector;
	}

	dq_Words.push_back (_pWord);
	b_LastWordIsNotInSection = _bWordIsNotInSection;
	if (true == _pWord->b_IsContentWord)
		hmp_TitleStems [_pWord->s_Stem] = true;
}


//-----------------------------------------------------------
bool Path::HasWord (Word& _rWord)
{
	return (hmp_TitleStems.find (_rWord.s_Stem) != hmp_TitleStems.end ());
}


//-----------------------------------------------------------
bool Path::operator< (Path& _rRight)
{
	return (d_Score < _rRight.d_Score);
}


//-----------------------------------------------------------
bool Path::IsYGood (Path& _rYGood) 
{
    if (!b_YGoodMatch)
		return false;

    if ((true == b_LastWordIsNotInSection) ||
		 (true == _rYGood.b_LastWordIsNotInSection))
	return true;

    if (dq_Words.back ()->s_Word != _rYGood.dq_Words.back ()->s_Word)  
	{
		b_YGoodMatch = false;
		return false;
    }

    return true;
}


//-----------------------------------------------------------
void Path::GetSectionAverageFeatures (FeatureVector& _rFeatureVector)
{
    for (int i = AVERAGE_FEATURE_OFFSET; 
		i < AVERAGE_FEATURE_OFFSET + BASE_FEATURE_COUNT; 
		++ i) 
	{
		if (arr_SectionFeatureCounts [i] > 0)
			_rFeatureVector.arr_Features [i] = arr_SectionFeatureTotals [i] 
											/ (double) arr_SectionFeatureCounts [i];
    }
}


//-----------------------------------------------------------
void Path::AccumulateSectionFeatures (void)
{
    if (NULL == p_FeatureVector)
		return;
    
    for (int i = 0; i < BASE_FEATURE_COUNT; ++ i) 
	{
		arr_SectionFeatureTotals [AVERAGE_FEATURE_OFFSET + i] 
								+= p_FeatureVector->arr_Features [i];
		++ arr_SectionFeatureCounts [AVERAGE_FEATURE_OFFSET + i];
    }
}

//-----------------------------------------------------------
int Path::GetSimilarity (Path* pPath)
{
    int iSimilarity = 0;

    Unigram_set_t setPathWords;
    ITERATE_COLLECTION (Word_dq_t, dq_Words, iteWord) 
	{
		Word* pWord = *iteWord;
		if (pWord->b_IsContentWord)
			// setPathWords.insert (pWord->s_Stem);
			setPathWords.insert (pWord->s_Word);
    }
    
    ITERATE_COLLECTION (Word_dq_t, pPath->dq_Words, itePathWord) 
	{
		Word* pWord = *itePathWord;
		if (pWord->b_IsContentWord)
			// if (setPathWords.find (pWord->s_Stem) != setPathWords.end())
			if (setPathWords.find (pWord->s_Word) != setPathWords.end())
				++ iSimilarity;
    }

    return iSimilarity;
}

//-----------------------------------------------------------
ostream& operator<< (ostream& _ofs, const Path& _rPath)
{
	_ofs << _rPath.d_Score << "   ";
	ITERATE_COLLECTION (Word_dq_t, ((Path&)_rPath).dq_Words, ite)
	{
		Word* pWord = *ite;
		if (false == pWord->b_StartWord)
			_ofs << pWord->s_Word << '/' << pWord->s_POSTag << ' ';
		else
			_ofs << "> ";
	}

	return _ofs;
}


//-----------------------------------------------------------
PathStore::PathStore (Path* _pPath, Word* _pWord, FeatureVector* _pFeatureVector, double _dScore)
{
	p_Path = _pPath;
	p_Word = _pWord;
	p_FeatureVector = _pFeatureVector;
	d_Score = _dScore;
}

PathStore::~PathStore (void)
{
	if (NULL != p_FeatureVector)
		delete p_FeatureVector;
}

//-----------------------------------------------------------
bool PathStore::operator< (const PathStore& _rRight) const
{
	return (d_Score < _rRight.d_Score);
}


//-----------------------------------------------------------
Beam::Beam (bool _bTrain)
{
	int iWindowSize = Config::Get ("WINDOW_SIZE");
	Path::i_InitialStartWords = iWindowSize - 1;

	i_MaxBeamWidth = _bTrain ? Config::Get ("TRAIN_BEAM_SIZE") : Config::Get ("DECODE_BEAM_SIZE");
	pp_Paths = new Path* [i_MaxBeamWidth];
	memset (pp_Paths, 0, i_MaxBeamWidth * sizeof (Path*));

	i_BeamWidth = 1;
	pp_Paths [0] = new Path;
}


//-----------------------------------------------------------
Beam::Beam (Beam &_rOld, bool _bCollapse)
{
    i_MaxBeamWidth = _rOld.i_MaxBeamWidth;
    i_BeamWidth = _rOld.i_BeamWidth;
    d_CurrentMin = _rOld.d_CurrentMin;

    pp_Paths = new Path* [i_MaxBeamWidth];
    memset (pp_Paths, 0, i_MaxBeamWidth * sizeof (Path*));

    if (_bCollapse) 
	{
		set<String> setPathTitles;    
		int new_i = 0;
		for (int old_i = 0; old_i < _rOld.i_BeamWidth; ++ old_i) 
		{
			String sBeamTitle;
			Utils::WordListToNgramString ( sBeamTitle, &(_rOld.pp_Paths[old_i]->dq_Words), NULL);
			
			if (setPathTitles.find( sBeamTitle) == setPathTitles.end()) 
			{
				setPathTitles.insert (sBeamTitle);
				pp_Paths [new_i] = new Path (*_rOld.pp_Paths [old_i]);
				pp_Paths [new_i]->d_Score = _rOld.i_BeamWidth - old_i;
				++ new_i;
			}
		}
		i_BeamWidth = new_i;

	}
	else
	{
		for (int i = 0; i < _rOld.i_BeamWidth; ++ i) 
			pp_Paths [i] = new Path (*_rOld.pp_Paths [i]);
    }
}


//-----------------------------------------------------------
Beam::~Beam (void)
{
	for (int i = 0; i < i_BeamWidth; ++ i)
		delete pp_Paths [i];

	i_BeamWidth = -1;
	i_MaxBeamWidth = -2;
	delete[] pp_Paths;
}


//-----------------------------------------------------------
void Beam::Clear (void) 
{
	for (int i = 0; i < i_BeamWidth; ++ i)
		delete pp_Paths [i];

	i_BeamWidth = 0;
}


//-----------------------------------------------------------
void Beam::Init ( void) 
{
	for (int i = 0; i < i_BeamWidth; ++ i)
		delete pp_Paths [i];

	i_BeamWidth = 1;
	pp_Paths [0] = new Path;
}


//-----------------------------------------------------------
Path* Beam::operator[] (int _iIndex)
{
	assert (_iIndex < i_BeamWidth);
	return pp_Paths [_iIndex];
}


//-----------------------------------------------------------
void Beam::AddPath (Path* _pPath, Word* _pWord, FeatureVector* _pFeatureVector, double _dScore)
{
	if ((map_AllNewPaths.size () == (unsigned int) i_MaxBeamWidth) && (_dScore < d_CurrentMin))
	{
		delete _pFeatureVector;
	 	return;
	}

	map_AllNewPaths.insert (make_pair (_dScore, new PathStore (_pPath, _pWord, _pFeatureVector, _dScore)));
	if (map_AllNewPaths.size () > (unsigned int) i_MaxBeamWidth)
	{
		Score_to_PathWord_map_t::iterator	ite = map_AllNewPaths.begin ();
		delete ite->second;
		map_AllNewPaths.erase (ite);
	}

	Score_to_PathWord_map_t::iterator ite = map_AllNewPaths.begin ();
	d_CurrentMin = ite->first;
}


//-----------------------------------------------------------
void Beam::Trim (void)
{
	Path_map_t	mapPaths;
	for (int i = 0; i < i_BeamWidth; ++ i)
		mapPaths.insert (make_pair (pp_Paths [i], pp_Paths [i]));

	int iPathCount = 0;
	Score_to_PathWord_map_t::reverse_iterator	ite;
	for (ite = map_AllNewPaths.rbegin ();
		ite != map_AllNewPaths.rend ();
		++ ite)
	{
		PathStore* pStore = ite->second;
		Path* pPath = pStore->p_Path;

		Path_map_t::iterator itePath;
		itePath = mapPaths.find (pPath);
		if (itePath != mapPaths.end ())
		{
			pPath = itePath->second;
			itePath->second = new Path (*pPath);
		}

		pPath->Grow (pStore->p_Word, pStore->d_Score, pStore->p_FeatureVector);
		pPath->AccumulateSectionFeatures ();
		pStore->p_FeatureVector = NULL;
		pp_Paths [iPathCount ++] = pPath;

		if (iPathCount >= i_MaxBeamWidth)
			break;
	}
	// delete path stores...			
	for (ite = map_AllNewPaths.rbegin ();
		ite != map_AllNewPaths.rend ();
		++ ite)
	{
		delete ite->second;
	}
	map_AllNewPaths.clear ();

	i_BeamWidth = iPathCount;

	Path_map_t::iterator	itePath;
	for (itePath = mapPaths.begin ();
		itePath != mapPaths.end ();
		++ itePath)
		delete itePath->second;
	mapPaths.clear ();
}


//-----------------------------------------------------------
bool Beam::HasYGood (Path& _rYGood)
{
	for (int i = 0; i < i_BeamWidth; ++ i)
	{
	    if ((*pp_Paths [i]).IsYGood (_rYGood))
		return true;
	}
	return false;
}


//-----------------------------------------------------------
void Beam::Reset (Path& _rYGood)
{
	for (int i = 0; i < i_BeamWidth; ++ i)
		delete pp_Paths [i];

	pp_Paths [0] = new Path (_rYGood);
	i_BeamWidth = 1;
}

//-----------------------------------------------------------
Decoder::Decoder (Model& _rModel, bool _bTrain)
	: beam (_bTrain)
{
    p_Model = &_rModel;
	p_YGood = NULL;
    Path::p_StartWord = Word::GetStartWord ();
}


//-----------------------------------------------------------
Decoder::~Decoder (void)
{
	if (NULL != p_YGood)
		delete p_YGood;
}


//-----------------------------------------------------------
String Decoder::BeamSearch (Section* _pSection, bool _bTrain, bool _bFlat)
{
	Perceptron& rPerceptron = p_Model->o_TitlePerceptron;


	beam.Clear ();
	if (NULL != p_YGood)
		delete p_YGood;
	p_YGood = new Path;

	// Get config options
	String sWeightsFileName (Config::Get ("WEIGHTS_FILE_PATH"));
	int iBeamWidthToPrint = Config::Get ("BEAM_WIDTH_TO_PRINT");
	bool bUseOneBest = (1 == (int) Config::Get("DECODER_USE_ONE_BEST"));
	
	// Initialize with the #start# words in front is done 	
	// automatically by the path constructor
	Word* pEndWord = Word::GetEndWord ();
	FeatureVector::Init (*p_Model, _pSection);

	p_YGood->p_FeatureVector = new FeatureVector;

	// Get Title Length
	int iPathLength = Config::Get ("TITLE_LENGTH");
	if (true == _bTrain)
	    iPathLength = _pSection->p_Title->lst_Words.size ();
	else
	{
		if (true == _pSection->p_Title->lst_Words.empty ())
			return String ();
		if (0 == iPathLength)
			iPathLength = Utils::GetTitleWordLength (_pSection->p_Title->s_Text) - 1;
	}
	if (iPathLength <= 0)
		return String ();

	cout << endl <<  "     Search section: " << _pSection->p_Title->s_Text << endl;

	Word_dq_t* pdqWords = &_pSection->dq_UniqueWords;
	if (true == pdqWords->empty ())
		return String ();
	Word_list_t* plstTitleWords = &_pSection->p_Title->lst_Words;
	if (true == plstTitleWords->empty ()) 
		return String ();

	beam.Init ();

	Word_list_t::iterator	iteTitleWord = plstTitleWords->begin ();
	int iTitleLength = plstTitleWords->size ();
	for (int iLength = 0; iLength < iPathLength - 1; ++ iLength) 
	{
	    // Grow Y Good
		if (true == _bTrain)
		{
			Word* pWord = *iteTitleWord;
			bool bSpecialWord = (0 == _pSection->GetUnigramOccurrence (pWord->s_Word, false));

			p_YGood->p_FeatureVector->Clear ();
			p_YGood->p_FeatureVector->CreateFeatures (p_YGood, pWord);
			p_YGood->GetSectionAverageFeatures (*p_YGood->p_FeatureVector);
			p_YGood->Grow (pWord, 0, p_YGood->p_FeatureVector, bSpecialWord);		
	    }
		else if (iLength < iTitleLength - 1)
	    {
			Word* pWord = *iteTitleWord;
			bool bSpecialWord = (0 == _pSection->GetUnigramOccurrence (pWord->s_Word, false));
			p_YGood->Grow (pWord, 0, p_YGood->p_FeatureVector, bSpecialWord);		
	    }

	    // Grow beam
	    for (int iPath = 0; iPath < beam.i_BeamWidth; ++ iPath) 
		{
			Path* pPath = beam [iPath];
		
			ITERATE_COLLECTION_POINTER ( Word_dq_t, pdqWords, iteWord) 
			{
				Word* pWord = *iteWord;
				if (true == pPath->HasWord (*pWord))
					continue;
				
				FeatureVector* pFV = new FeatureVector;
				pFV->CreateFeatures (pPath, pWord);
				pPath->GetSectionAverageFeatures (*pFV);
				double dScore = rPerceptron.Evaluate (*pFV);
				
				beam.AddPath (pPath, pWord, pFV, dScore);
			}
	    }

	    // Trim
	    beam.Trim ();

	    // Print out beam
	    if (false == _bTrain) 
		{
			for (int i = 0; (i < iBeamWidthToPrint) && (i < beam.i_BeamWidth); ++ i)
				cout << "   " << i << "  " << *beam [i] << endl;
	    }

	    // Perform Y Good in beam update
	    p_YGood->AccumulateSectionFeatures ();

	    if (true == _bTrain) 
		{
			if (false == beam.HasYGood (*p_YGood)) 
			{
				cout << endl;
				cout << "     -- Y Good : " << *p_YGood << endl;
				for (int iBeam = 0; (iBeam < beam.i_BeamWidth) && (iBeam < 2); ++ iBeam) 
				{
					Path* pPath = beam [iBeam];
					cout << "     " << *pPath << endl;
				}

				Feature_Vector_list_t lstFVBad;
				for (int iBeam = 0; iBeam < beam.i_BeamWidth; ++ iBeam) 
				{
					Path* pPath = beam [iBeam];
					lstFVBad.push_back (pPath->p_FeatureVector);
				}
		    
				rPerceptron.Update (*p_YGood->p_FeatureVector, lstFVBad);
				beam.Reset (*p_YGood);

			}
			else
			{
				cout << '.';
				cout.flush ();
			}
	    }
	    if (iLength < iTitleLength - 1)
		    ++ iteTitleWord;
	}

	// Last iteration for the last word and end word

	// Grow Y Good
	if (true == _bTrain)
	{
	    Word* pWord = *iteTitleWord;
	    bool bSpecialWord = (0 == _pSection->GetUnigramOccurrence (pWord->s_Word, false));
	    
	    p_YGood->p_FeatureVector->Clear ();
	    // features for the last word...
	    p_YGood->p_FeatureVector->CreateFeatures (p_YGood, pWord);
	    p_YGood->Grow (pWord, 0, p_YGood->p_FeatureVector, bSpecialWord);
	    // features for the #end# word...
	    p_YGood->p_FeatureVector->CreateFeatures (p_YGood, pEndWord);
	    p_YGood->GetSectionAverageFeatures (*p_YGood->p_FeatureVector);
	}
	else
	{
	    Word* pWord = *iteTitleWord;
	    bool bSpecialWord = (0 == _pSection->GetUnigramOccurrence (pWord->s_Word, false));
	    p_YGood->Grow (pWord, 0, p_YGood->p_FeatureVector, bSpecialWord);
	}
	
	// Grow Beam
	// evalutation loop for the last word in the path	
	for (int iPath = 0; iPath < beam.i_BeamWidth; ++ iPath) 
	{
	    Path* pPath = beam [iPath];
	    
	    ITERATE_COLLECTION_POINTER ( Word_dq_t, pdqWords, iteWord) 
		{
			Word* pWord = *iteWord;
			if (true == pPath->HasWord (*pWord))
				continue;

			FeatureVector* pFV = new FeatureVector;
			pFV->CreateFeatures (pPath, pWord);

			Path oTempPath (*pPath);
			oTempPath.Grow (pWord, 0, NULL);
			pFV->CreateFeatures (&oTempPath, pEndWord);
			oTempPath.GetSectionAverageFeatures (*pFV);

			double dScore = rPerceptron.Evaluate (*pFV);
			
			beam.AddPath (pPath, pWord, pFV, dScore);
		}
	}

	beam.Trim ();

	// Print out beam
	if (false == _bTrain) 
	{
	    for (int i = 0; (i < iBeamWidthToPrint) && (i < beam.i_BeamWidth); ++ i)
			cout << "   " << i << "  " << *beam [i] << endl;
	}

	p_YGood->AccumulateSectionFeatures ();
	if (true ==  _bTrain) 
	{
		// If the Y Good is not the best in the beam, update
		// bool bUpdateNotRequired = (true == _bFlat)? 
		//								beam [0]->IsYGood (*p_YGood) : 
		//								beam.HasYGood (*p_YGood);
	    if (((false == bUseOneBest) && (false == beam.HasYGood (*p_YGood))) ||
			((true == bUseOneBest) && (false == beam [0]->IsYGood (*p_YGood))))
		{
			// Print out Y good
			cout << endl;
			cout << "     == Y Good : " << *p_YGood << endl;
			for (int iBeam = 0; (iBeam < beam.i_BeamWidth) && (iBeam < 2); ++ iBeam) 
			{
				Path* pPath = beam [iBeam];
				cout << "     " << *pPath << endl;
			}
		
			// Get bad paths with score high than Y good
			Feature_Vector_list_t lstFVBad;
			for (int iBeam = 0; iBeam < beam.i_BeamWidth; ++ iBeam) 
			{
				Path* pPath = beam [iBeam];
				if (pPath->IsYGood (*p_YGood))
					break;
			
				lstFVBad.push_back (pPath->p_FeatureVector);
			}

			// Update
			rPerceptron.Update (*p_YGood->p_FeatureVector, lstFVBad);
			if ( true == _bFlat) 
			    beam.Reset (*p_YGood);
		}


		// when we get here, beam [0] is the highest scoring path...	

		// for (int i = 0; i < beam.i_BeamWidth; ++ i)
		// 	beam [i]->IsYGood (*p_YGood);

		// debug >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
		int iBadCount = 0;
		for (int i = 0; i < beam.i_BeamWidth; ++ i)
		{
			if (false == beam [i]->IsYGood (*p_YGood))
				++ iBadCount;
		}
		if (iBadCount > 0)
			cout << '.' << iBadCount << '/' << beam.i_BeamWidth;
		// debug <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	
	    cout << endl;
	}
	else
	{
		cout << "-----------------------------------------------" << endl;
		if (iBeamWidthToPrint > beam.i_BeamWidth)
			iBeamWidthToPrint = beam.i_BeamWidth;
		for (int i = 0; i < iBeamWidthToPrint; ++ i)
			cout << "   >> " << *beam [i] << endl;
	}

	// construct the title to be returned...										
	String sTitle;
	if (false == _bTrain)
	{
		ITERATE_COLLECTION (Word_dq_t, beam [0]->dq_Words, iteWord)
		{
			Word* pWord = *iteWord;
			if ((false == pWord->b_StartWord) && (false == pWord->b_EndWord))
				sTitle << pWord->s_Word << ' ';
		}
	}

	// During TOC decoding, these are accessed by the TOC BeamSearch, and should	
	// not be cleared out - the Decoder destructor will do that.					
	// If we are doing flat decoding, clear out the internal search structures.		
	// This is done to catch inadvertent access of these structures by the TOC		
	// decoder during flat decoding...												
	if (true == _bFlat)
	{
		beam.Clear ();
		if (NULL != p_YGood)
			delete p_YGood;
		p_YGood = NULL;
	}

	return sTitle;
}

