#include <iostream>
#include <fstream>
#include <algorithm>
#include "LanguageModel.h"
#include "Config.h"
using namespace std;




//-----------------------------------------------------------------------------------------
LanguageModel::LanguageModel (void)
{
    i_Order = 0;
	p_StartWord = Word::GetStartWord ();
	p_EndWord = Word::GetEndWord ();
}

//-----------------------------------------------------------------------------------------
LanguageModel::~LanguageModel (void)
{
}

//-----------------------------------------------------------------------------------------
void LanguageModel::Train (const char* _zInputFilePath, const char* _zArpaFilePath, 
							int _iOrder, const char* _zSRILMOptions)
{
    // Train with SRILM
    String sSRILMCommandLine;
    sSRILMCommandLine << Config::Get ("SRILM_PATH") << "ngram-count "
                      << " -order " << _iOrder 
                      << " -text " << _zInputFilePath
                      << " -lm " << _zArpaFilePath
					  << " " << _zSRILMOptions;
    cerr << sSRILMCommandLine << endl;
    system (sSRILMCommandLine);

    // Load ARPA File
    LoadARPAFile (_zArpaFilePath);
}



//-----------------------------------------------------------------------------------------
void LanguageModel::LoadARPAFile (const char* _zFileName)
{
    ifstream file (_zFileName);
    if (false == file.is_open ())
	{
		cerr << "Failed to open file '" << _zFileName << "'." << endl;
		throw;
	}

	Word* pStartWord = Word::GetStartWord ();
	Word* pEndWord = Word::GetEndWord ();

    while (false == file.eof ())
	{
	    String sLine;
	    getline (file, sLine);
	 
	    sLine.Replace ("<s>", pStartWord->s_Word);
	    sLine.Replace ("</s>", pEndWord->s_Word);
	    
	    list<String> lstLineParts;
	    sLine.Split (lstLineParts, '\t');
	    
	    // Skip the lines that do not have ngrams and weights
	    if (lstLineParts.size () < 2)
			continue;

	    // Store ngram probability
	    list<String>::iterator iteLineParts = lstLineParts.begin ();
	    
	    double fProbability = *iteLineParts;
	    ++iteLineParts;
	    
	    String sNgram = *iteLineParts;
	    ++iteLineParts;

	    // Set Probability
	    map_Score_by_Ngram[ sNgram ] = fProbability;

	    // Set Backoff weight
	    if (iteLineParts != lstLineParts.end ())
			map_BackoffWeight_by_Ngram[ sNgram ] = *iteLineParts;

	    // Calculate order
	    list<String> lstNgramParts;
	    sNgram.Split (lstNgramParts);
	    if (lstNgramParts.size () > (size_t) i_Order)
			i_Order = lstNgramParts.size ();	 
	}
    file.close ();
}


//-----------------------------------------------------------------------------------------
#define UNKNOWN_PROB			-100
#define UNKNOWN_UNCERTAINTY		-100


//-----------------------------------------------------------------------------------------
pair <double, double> LanguageModel::Get (Word_dq_t& _dqWords, 
										  Word* _pNextWord, 
										  WordRenderer_t _pRenderer,
										  bool _bIgnoreTerminals)
{
    long iLength = _dqWords.size ();

	long iStart;
	if (false == _bIgnoreTerminals)
	{
		iStart = iLength - 1;
		for (long i = 0; i < iLength - 1; ++ i)
		{
			if (false == _dqWords [i + 1]->b_StartWord)
			{
				iStart = i;
				break;
			}
		}
	}
	else
	{
		iStart = iLength;
		for (long i = 0; i < iLength; ++ i)
		{
			if (false == _dqWords [i]->b_StartWord)
			{
				iStart = i;
				break;
			}
		}
	}
	
	// cout << "   length [" << iLength << "] - start [" << iStart << "] >= order [" << i_Order << ']' << endl;
	if (iLength - iStart >= (long) i_Order)
		return GetNGramProb (_dqWords, _pNextWord, iLength - i_Order + 1, i_Order - 1, _pRenderer);
	else
	{
		iLength -= iStart;
		return GetNGramProb (_dqWords, _pNextWord, iStart, iLength, _pRenderer);
	}
}


//-----------------------------------------------------------------------------------------
pair <double, double> LanguageModel::GetFullScore (Word_dq_t& _dqWords, 
													Word* _pNextWord,
													WordRenderer_t _pRenderer,
													bool _bIgnoreTerminals)
{
    long iLength = _dqWords.size ();

	long iStart;
	if (false == _bIgnoreTerminals)
	{
		iStart = iLength - 1;
		for (long i = 0; i < iLength - 1; ++ i)
		{
			if (false == _dqWords [i + 1]->b_StartWord)
			{
				iStart = i;
				break;
			}
		}
	}
	else
	{
		iStart = iLength;
		for (long i = 0; i < iLength; ++ i)
		{
			if (false == _dqWords [i]->b_StartWord)
			{
				iStart = i;
				break;
			}
		}
	}
	
	if (iLength - iStart >= (long) i_Order)
	{
		long lNormalizingFactor = 1;
		double dProbability = 0;
		double dUncertainty = 0;
		pair <double, double> pairScore;
		for (long i = iStart; i <= iLength - (long) i_Order; ++ i)
		{
			pairScore = GetNGramProb (_dqWords, _dqWords [i + i_Order - 1], i, i_Order - 1, _pRenderer, _bIgnoreTerminals);
			dProbability += pairScore.first;
			dUncertainty += pairScore.second;
			++ lNormalizingFactor;
		}
		
		pairScore = GetNGramProb (_dqWords, _pNextWord, iLength - i_Order + 1, i_Order - 1, _pRenderer, _bIgnoreTerminals);
		dProbability += pairScore.first;
		dUncertainty += pairScore.second;
		return make_pair (dProbability / (double) lNormalizingFactor, dUncertainty / (double) lNormalizingFactor);
	}
	else
		return GetNGramProb (_dqWords, _pNextWord, iStart, iLength - iStart, _pRenderer, _bIgnoreTerminals);
}


//-----------------------------------------------------------------------------------------
pair <double, double> LanguageModel::GetNGramProb (Word_dq_t& _dqWords, 
													Word* _pNextWord,
													int _iStart, int _iLength, 
													WordRenderer_t _pRenderer,
													bool _bIgnoreTerminals)
{   
	// cout << "   LM::GNGP (start [" << _iStart << "], length [" << _iLength << "])" << endl;
	if ((true == _bIgnoreTerminals) || (0 == _iLength))
	{
		while ((_iLength > 0) && (true == _dqWords [_iStart]->b_StartWord))
		{
			++ _iStart;
			-- _iLength;
		}
		while ((_iLength > 0) && (true == _dqWords [_iStart]->b_EndWord))
			-- _iLength;
	}


    // 1-GRAM: Return the score P (A)
    if (0 == _iLength)
	{
		Weight_by_Ngram_map_t::iterator  iteScore;
		iteScore = map_Score_by_Ngram.find (_pRenderer (*_pNextWord));
		if (iteScore != map_Score_by_Ngram.end ())
		{
			// cout << "   P(" << _pRenderer (*_pNextWord) << ") = " << iteScore->second << endl;
			return make_pair (iteScore->second, 0);
		}
		else
		{
			iteScore = map_Score_by_Ngram.find ("<unk>");
			if (iteScore != map_Score_by_Ngram.end ())
			{
				// cout << "   P(" << _pRenderer (*_pNextWord) << ") = " << iteScore->second << endl;
				return make_pair (iteScore->second, -1);
			}

			cerr << "Probability not available for unigram '" << _pRenderer (*_pNextWord)
				 << "' and no <unk> token present in the language model !" << endl;
			assert (false);
			return make_pair (UNKNOWN_PROB, UNKNOWN_UNCERTAINTY);
		}
    }
    
    // 2-GRAM (AB): Return score P (B|A), interpolate if not found 
    if (1 == _iLength)
	{
		String sBigram;
		sBigram << _pRenderer (*_dqWords [_iStart]) << ' ' << _pRenderer (*_pNextWord);

		Weight_by_Ngram_map_t::iterator  iteScore;
		iteScore = map_Score_by_Ngram.find (sBigram);
		if (iteScore != map_Score_by_Ngram.end ())
		{
			// cout << "   P(" << sBigram << ") = " << iteScore->second << endl;
			return make_pair (iteScore->second, 0);
		}

		else 
		{
			// Interpolate P (B) + B (A)
			pair<double, double> pairScoreB = GetNGramProb (_dqWords, _pNextWord, _iStart + 1, 0, _pRenderer);

			double dBackoffA = 0;
			Weight_by_Ngram_map_t::iterator  iteBackoff;
			iteBackoff = map_BackoffWeight_by_Ngram.find (_pRenderer (*_dqWords [_iStart]));
			if (iteBackoff != map_BackoffWeight_by_Ngram.end ())
				dBackoffA = iteBackoff->second;
			
			// cout << "   B(" << _pRenderer (*_dqWords [_iStart]) << ") = " << dBackoffA << endl;
			map_Score_by_Ngram.insert (make_pair (sBigram, pairScoreB.first + dBackoffA));
			return make_pair (pairScoreB.first + dBackoffA, pairScoreB.second - 1);
		}
    }

    // 3-GRAM (ABC): Return score P (C|B,A), interpolate if not found
    if (2 == _iLength)
	{
		String sTrigram;
		sTrigram << _pRenderer (*_dqWords [_iStart]) << ' ' 
				 << _pRenderer (*_dqWords [_iStart + 1]) << ' ' 
				 << _pRenderer (*_pNextWord);

		Weight_by_Ngram_map_t::iterator  iteScore;
		iteScore = map_Score_by_Ngram.find (sTrigram);
		if (iteScore != map_Score_by_Ngram.end ())
		{
			// cout << "   P(" << sTrigram << ") = " << iteScore->second << endl;
			return make_pair (iteScore->second, 0);
		}

		else 
		{
			// Interpolate P (C|B) + B (B|A)
			// P (C|B)
			pair<double, double> pairScoreBC = GetNGramProb (_dqWords, _pNextWord, _iStart + 1, 1, _pRenderer);
	    
			// B (B|A)
			String sAB;
			sAB << _pRenderer (*_dqWords [_iStart]) << ' ' << _pRenderer (*_dqWords [_iStart + 1]);

			double dBackoffAB = 0;
			Weight_by_Ngram_map_t::iterator  iteBackoff;
			iteBackoff = map_BackoffWeight_by_Ngram.find (sAB);
			if (iteBackoff != map_BackoffWeight_by_Ngram.end ())
				dBackoffAB = iteBackoff->second;

			// cout << "   B(" << sAB << ") = " << dBackoffAB << endl;
			map_Score_by_Ngram.insert (make_pair (sTrigram, pairScoreBC.first + dBackoffAB));
			return make_pair (pairScoreBC.first + dBackoffAB, pairScoreBC.second - 1);
		}
    }
    
	cerr << "Invalid code path! LanguageModel::GetNGramProb () should not execute this line. "
		 << " Language model order = " << i_Order 
		 << " NGram length = " << _iLength << "." << endl;
    exit (1);
}



