#include "Config.h"
#include "Section.h"
#include "BookSection.h"
#include "Tagger.h"
#include "Reader.h"
#include "Decoder.h"
#include "TreeDecoder.h"
#include "LanguageModel.h"
#include "Utils.h"
#include "FeatureEnumerator.h"
#include "nlp_string.h"
#include <sys/time.h>
#include <time.h>



void TestTagLM (BookSection* _pRootSection);
void ComputeStats (Section* _pRootSection, int _iStartDepth);

void TagFromFile (BookSection* _pRootSection);
void Preprocess (BookSection* _pRoot);
void PreprocessCorpus (Model& _model, 
						BookSection* _pRootSection, 
						BookSection_list_t* _plstSections = NULL);
void Preprocess (Section* _pRoot);
void PreprocessCorpus (Model& _model, 
						Section* _pRootSection, 
						Section_list_t* _plstSections = NULL);

void RandomizeCorpus (Section* _pRoot);
void GetTrainingSections (Section* _pRoot, Section_list_t& _lstSections);
void GetDecodingSections (Section* _pRoot, Section_list_t& _lstSections);

void TrainFlat (Model& _rModel, Section* _pRoot, Section_list_t& _lstSections);
void DecodeFlat (Model& _rModel, Section* _pRoot, Section_list_t& _lstSections);
void TrainTOC (Model& _rModel, Section* _pRoot, Section_list_t& _lstSections);
void DecodeTOC (Model& _rModel, Section* _pRoot, Section_list_t& _lstSections);


//--------------------------------------------------------------------------------
void PrintTableOfContents (Section_dq_t dqTrainingSections, 
							Section_to_Title_map_t mapSectionToTitle, 
							bool _bTOC) 
{
	int iMinDepth = 100;
	ITERATE_COLLECTION (Section_dq_t, dqTrainingSections, ite)
	{
		Section* pSection = *ite;
		int iDepth = pSection->GetDepth ();
		if (iMinDepth >= iDepth)
			iMinDepth = iDepth;
	}

	ITERATE_COLLECTION (Section_dq_t, dqTrainingSections, iteSection)
	{
		Section* pSection = *iteSection;
		int iDepth = pSection->GetDepth ();
		if (iDepth == iMinDepth)
			cout << "== SUB TREE =================================================" 
				 << endl;

		for (int i=0; i < iDepth; ++i)
			cout << "    ";
		if (true == _bTOC)
			cout << "TOC : " << mapSectionToTitle [(long) pSection] << endl;
		else
			cout << "FLT : " << mapSectionToTitle [(long) pSection] << endl;

		for (int i=0; i < iDepth; ++i)
			cout << "    ";
		cout << "ACT : " << pSection->p_Title->s_Text << endl;
		cout << endl;
	}
}



//--------------------------------------------------------------------------------
int main (int argc, const char* argv[])
{
    assert (argc > 1);

    Config::SetCommandLine (argc, argv);

    // Initialize Feature Enumerator ----------
	FeatureEnumerator::Init ();

    // Read corpus ----------------------------
    cout << "* Reading corpus..." << endl;
    String sDataset = Config::Get ("DATASET");
    String sMode = Config::Get ("MODE");
    Section* pRootSection;

	// ENCODE CORPUS TO FEATURE XML ---------------------------------------
	if (sMode == "encode_xml")
	{
		Reader* pReader;
		pReader = Reader::CreateReader (Config::Get ("DATASET"));
		BookSection* pRootSection = pReader->ParseFile (Config::Get ("DATASET_PATH"));
		Model model;
		PreprocessCorpus (model, pRootSection);
		delete pReader;

		String sXML = BookSection::GetXML ();

		String sXMLFileName = Config::Get ("XML_FILENAME");
		if (sXMLFileName == "")
			return 1;

		ofstream file (sXMLFileName);
		if (false == file.is_open ())
			return 1;
		file << sXML << endl;
		file.close ();
		return 0;
	}
	
    // Tag ------------------------------------
    if (sMode == "tag") 
	{
		Reader* pReader;
		pReader = Reader::CreateReader (Config::Get ("DATASET"));
		BookSection* pRootSection = pReader->ParseFile (Config::Get ("DATASET_PATH"));
		delete pReader;
		Tagger tagger;
		tagger.TagCorpus (pRootSection);
		return 0;
    }
    
	// LM Test mode -------------------------------------------------------
	if (sMode == "test_lm")
	{
		Reader* pReader;
		pReader = Reader::CreateReader (Config::Get ("DATASET"));
		BookSection* pRootSection = pReader->ParseFile (Config::Get ("DATASET_PATH"));
		delete pReader;

		TestTagLM (pRootSection);
		return 0;
	}

	// Calculate statistics  ----------------------------------------------
	if (sMode == "stats")
	{
	    ComputeStats( pRootSection, Config::Get ("TOC_SECTION_DEPTH") );
	    return 0;
	}
	
	// PREPROCESS ---------------------------------------------------------
    if (sMode == "preprocess") 
	{
		Reader* pReader;
		pReader = Reader::CreateReader (Config::Get ("DATASET"));
		BookSection* pRootSection = pReader->ParseFile (Config::Get ("DATASET_PATH"));
		delete pReader;
		Preprocess (pRootSection);
		return 0;
	}

	// RANDOMIZE ----------------------------------------------------------
    if (sMode == "randomize") 
	{
		cout << "------------------------------------------------"  << endl;
		cout << "  CORPUS RANDOMIZATION MODE REQUESTED" << endl;
		cout << "------------------------------------------------"  << endl;
		cout << "  This mode randomizes the training & testing" << endl;
		cout << "  sections of the corpus.  Doing so will make" << endl;
		cout << "  all language models, perceptron weights and" << endl;
		cout << "  decoding results from previous runs INvALID." << endl;
		cout << "------------------------------------------------"  << endl;
		cout << "  Do you wish to continue? [Yes/n] :";
		cout.flush ();
		char zResponse [1000];
		cin >> zResponse;
		if (0 != strcmp (zResponse, "Yes"))
		{
			cout << "------------------------------------------------"  << endl;
			cout << "  Corpus not modified." << endl;
			cout << "------------------------------------------------"  << endl;
			return 0;
		}

		RandomizeCorpus (pRootSection);

		cout << "------------------------------------------------"  << endl;
		cout << "  CORPUS RANDOMIZED." << endl;
		cout << "------------------------------------------------"  << endl;

		return 0;
	}

	XMLReader* pReader = new XMLReader ();
	pRootSection = pReader->ParseFile (Config::Get ("XML_FILENAME"));
	delete pReader;

	// hack
	{
		Ngram_list_t lstNgramList;
		Word_dq_t dqWords;
		Utils::ExtractNounPhrases (lstNgramList, dqWords);
	}

	// TRAINING -----------------------------------------------------------
    if (sMode == "train") 
	{
		Config::Set ("TYPE", "flat");

		Section_list_t	lstSubtrees;
		GetTrainingSections (pRootSection, lstSubtrees);

		Section_list_t	lstSections;
		ITERATE_COLLECTION (Section_list_t, lstSubtrees, ite)
		{
			Section* pSection = *ite;
			Utils::GetInOrderSectionList (lstSections, pSection);
		}

		Model model;
		cout << "* Preprocessing" << endl;
		// PreprocessCorpus (model, pRootSection, &lstSections);
		PreprocessCorpus (model, pRootSection);

		cout << "  - training on " << lstSections.size () << " sections." << endl;
		TrainFlat (model, pRootSection, lstSections);
		return 0;
	}

	// DECODING -----------------------------------------------------------
    if (sMode == "decode") 
	{
		Config::Set ("TYPE", "flat");

		Section_list_t	lstSubtrees;
		GetDecodingSections (pRootSection, lstSubtrees);

		Section_list_t	lstSections;
		ITERATE_COLLECTION (Section_list_t, lstSubtrees, ite)
		{
			Section* pSection = *ite;
			Utils::GetInOrderSectionList (lstSections, pSection);
		}

		Model model;
		cout << "* Preprocessing" << endl;
		PreprocessCorpus (model, pRootSection);

		cout << "  - decoding on " << lstSections.size () << " sections." << endl;
		DecodeFlat (model, pRootSection, lstSections);
		return 0;
	}

	// TRAINING FOR TOC ---------------------------------------------------
    if (sMode == "toc_train") 
	{
		Config::Set ("TYPE", "toc");

		Section_list_t	lstSubtrees;
		GetTrainingSections (pRootSection, lstSubtrees);

		Section_list_t	lstSections;
		ITERATE_COLLECTION (Section_list_t, lstSubtrees, ite)
		{
			Section* pSection = *ite;
			Utils::GetInOrderSectionList (lstSections, pSection);
		}

		Model model;
		cout << "* Preprocessing" << endl;
		// PreprocessCorpus (model, pRootSection, &lstSections);
		PreprocessCorpus (model, pRootSection);

		cout << "  - training on " << lstSubtrees.size () << " subtrees, " 
			 << lstSections.size () << " sections." << endl;
		TrainTOC (model, pRootSection, lstSubtrees);
		return 0;
	}

	// DECODING FOR TOC ---------------------------------------------------
    if (sMode == "toc_decode") 
	{
		Config::Set ("TYPE", "toc");

		Section_list_t	lstSubtrees;
		GetDecodingSections (pRootSection, lstSubtrees);

		Model model;
		cout << "* Preprocessing" << endl;
		PreprocessCorpus (model, pRootSection);

		cout << "  - decoding on " << lstSubtrees.size () << " subtrees." << endl;
		DecodeTOC (model, pRootSection, lstSubtrees);
		return 0;
	}
	
    cerr << "Invalid mode: " << sMode << endl; 

    return 1;
}
