#include "Config.h"
#include "BookSection.h"
#include "Tagger.h"
#include "Reader.h"
#include "Decoder.h"
#include "TreeDecoder.h"
#include "LanguageModel.h"
#include "Utils.h"
#include "FeatureEnumerator.h"
#include "nlp_string.h"
#include <sys/time.h>
#include <time.h>


void PreprocessCorpus (Model& _model, 
						Section* _pRootSection, 
						Section_list_t* _plstSections = NULL);
void PreprocessCorpus (Model& _model, 
						BookSection* _pRootSection, 
						BookSection_list_t* _plstSections = NULL);


//--------------------------------------------------------------------------------
void BuildTitleTagLM (LanguageModel &_lmTag, Section_list_t* _plstSections)
{
    String sInputFilePath = Config::GetTempFileName ("TAG_LM_INPUT_PATH");
    String sArpaFilePath = Config::Get ("TAG_LM_OUTPUT_PATH");
    String sOptions = Config::Get ("TAG_LM_SRILM_OPTIONS");
	int iOrder = Config::Get ("TAG_LM_ORDER");
    
	if (NULL == _plstSections)
	{
		if (0 == access (sArpaFilePath, R_OK))
		{
			cout << "* Loading title tag language model..." << endl;
			_lmTag.LoadARPAFile (sArpaFilePath);
			return;
		}
	}

    cout << "* Building title tag language model..." << endl;

    // Create language model input file
    ofstream file (sInputFilePath);

    ITERATE_COLLECTION_POINTER (Section_list_t, _plstSections, iteSection) 
	{
		Unigram_list_t lstTagList;
		Utils::WordListToTagList (lstTagList, & ((*iteSection)->p_Title->lst_Words));

		String sTitleTags;
		sTitleTags.Join (lstTagList);
		file << sTitleTags << endl;
    }
    file.close ();

    // Build lexical language model
    _lmTag.Train (sInputFilePath, sArpaFilePath, iOrder, sOptions);
	
    if (Config::Get ("RETAIN_TAGLM_INPUT_FILE") != 1)
		remove (sInputFilePath);
}




//--------------------------------------------------------------------------------
void BuildLexicalLM (Section* _pRootSection, 
					LanguageModel &_lmLexical, 
					Section_list_t* _plstSection)
{
    String sInputFilePath = Config::GetTempFileName ("LM_INPUT_PATH");
    String sArpaFilePath = Config::Get ("LM_OUTPUT_PATH");
    String sOptions = Config::Get ("LM_SRILM_OPTIONS");
	int iOrder = Config::Get ("LM_ORDER");
    
	if (NULL == _plstSection)
	{
		if (0 == access (sArpaFilePath, R_OK))
		{
			cout << "* Loading lexical language model..." << endl;
			_lmLexical.LoadARPAFile (sArpaFilePath);
			return;
		}
	}
    
    cout << "* Building Lexical Language Model..." << endl;

    // Create language model input file
	bool bUseWordAndTags = ( 1 == (int)Config::Get( "LM_USE_TAGS") );
    ofstream file (sInputFilePath);

	ITERATE_COLLECTION_POINTER (Section_list_t, _plstSection, iteTitles)
	{
		Section* pSection = *iteTitles;
		if( bUseWordAndTags ) 
			file << pSection->p_Title->GetTaggedText () << endl;
		else 
		{
			String sLowerText = pSection->p_Title->s_Text;
			sLowerText.LowerCase();
			file << sLowerText << endl;
		}
	}

    ITERATE_COLLECTION_POINTER (Section_list_t, _pRootSection->GetAllSections (), iteSection) 
	{
		Section* pSection = *iteSection;
		ITERATE_COLLECTION (Sentence_list_t, pSection->lst_Sentences, iteSentence) 
		{
		    if (bUseWordAndTags) 
				file << (*iteSentence)->GetTaggedText () << endl;
		    else 
			{
				String sLowerText = (*iteSentence)->s_Text;
				sLowerText.LowerCase();
				file << sLowerText << endl;
		    }
		}
    }
    file.close ();

    // Build lexical language model
    _lmLexical.Train (sInputFilePath, sArpaFilePath, iOrder, sOptions);
	
    if (Config::Get ("RETAIN_LM_INPUT_FILE") != 1)
		remove (sInputFilePath);
}



//--------------------------------------------------------------------------------
void BuildTitleTagLM (LanguageModel &_lmTag, BookSection_list_t* _plstSections)
{
    String sInputFilePath = Config::GetTempFileName ("TAG_LM_INPUT_PATH");
    String sArpaFilePath = Config::Get ("TAG_LM_OUTPUT_PATH");
    String sOptions = Config::Get ("TAG_LM_SRILM_OPTIONS");
	int iOrder = Config::Get ("TAG_LM_ORDER");
    
	if (NULL == _plstSections)
	{
		if (0 == access (sArpaFilePath, R_OK))
		{
			cout << "* Loading title tag language model..." << endl;
			_lmTag.LoadARPAFile (sArpaFilePath);
		}
		else
			cerr << "Language model file '" << sArpaFilePath 
				 << "' not found." << endl
				 << "This file is needed for the current MODE." << endl
				 << "Please restore this file, or re-run with MODE=preprocess "
				 << "on the command line to create this file." << endl;
		return;
	}

    cout << "* Building title tag language model..." << endl;

    // Create language model input file
    ofstream file (sInputFilePath);

    ITERATE_COLLECTION_POINTER (BookSection_list_t, _plstSections, iteSection) 
	{
		Unigram_list_t lstTagList;
		Utils::WordListToTagList (lstTagList, & ((*iteSection)->p_Title->lst_Words));

		String sTitleTags;
		sTitleTags.Join (lstTagList);
		file << sTitleTags << endl;
    }
    file.close ();

    // Build lexical language model
    _lmTag.Train (sInputFilePath, sArpaFilePath, iOrder, sOptions);
	
    if (Config::Get ("RETAIN_TAGLM_INPUT_FILE") != 1)
		remove (sInputFilePath);
}




//--------------------------------------------------------------------------------
void BuildLexicalLM (BookSection* _pRootSection, 
					LanguageModel &_lmLexical, 
					BookSection_list_t* _plstSection)
{
    String sInputFilePath = Config::GetTempFileName ("LM_INPUT_PATH");
    String sArpaFilePath = Config::Get ("LM_OUTPUT_PATH");
    String sOptions = Config::Get ("LM_SRILM_OPTIONS");
	int iOrder = Config::Get ("LM_ORDER");
    
	if (NULL == _plstSection)
	{
		if (0 == access (sArpaFilePath, R_OK))
		{
			cout << "* Loading lexical language model..." << endl;
			_lmLexical.LoadARPAFile (sArpaFilePath);
			return;
		}
	}
    
    cout << "* Building Lexical Language Model..." << endl;

    // Create language model input file
	bool bUseWordAndTags = ( 1 == (int)Config::Get( "LM_USE_TAGS") );
    ofstream file (sInputFilePath);

	ITERATE_COLLECTION_POINTER (BookSection_list_t, _plstSection, iteTitles)
	{
		BookSection* pSection = *iteTitles;
		if( bUseWordAndTags ) 
			file << pSection->p_Title->GetTaggedText () << endl;
		else 
		{
			String sLowerText = pSection->p_Title->s_Text;
			sLowerText.LowerCase();
			file << sLowerText << endl;
		}
	}

    ITERATE_COLLECTION_POINTER (BookSection_list_t, _pRootSection->GetAllSections (), iteSection) 
	{
		BookSection* pSection = *iteSection;
		ITERATE_COLLECTION (Sentence_list_t, pSection->lst_Sentences, iteSentence) 
		{
		    if (bUseWordAndTags) 
				file << (*iteSentence)->GetTaggedText () << endl;
		    else 
			{
				String sLowerText = (*iteSentence)->s_Text;
				sLowerText.LowerCase();
				file << sLowerText << endl;
		    }
		}
    }
    file.close ();

    // Build lexical language model
    _lmLexical.Train (sInputFilePath, sArpaFilePath, iOrder, sOptions);
	
    if (Config::Get ("RETAIN_LM_INPUT_FILE") != 1)
		remove (sInputFilePath);
}



//--------------------------------------------------------------------------------
void TestTagLM (BookSection* _pRootSection)
{
	cout << "* testing lm scores" << endl;
	String sFileName (Config::Get ("LM_TEST_FILE"));
	ifstream file (sFileName);
	if (false == file.is_open ())
	{
		cerr << "Failed to open file '" << sFileName << "'." << endl;
		return;
	}

	Model model;
	PreprocessCorpus (model, _pRootSection);

	while (false == file.eof ())
	{
		String sLine;
		getline (file, sLine);
		if ((true == file.eof ()) && (sLine == ""))
			break;

		Unigram_list_t	lstNGram;
		sLine.Split (lstNGram);
		
		Word_dq_t	dqWords;
		Unigram_list_t::iterator	ite;
		for (ite = lstNGram.begin ();
			ite != lstNGram.end ();
			++ ite)
		{
			Unigram_list_t	lstWord;
			ite->Split (lstWord, '_');

			Word* pWord = new Word (lstWord.front ());
			pWord->s_POSTag = lstWord.back ();
			dqWords.push_back (pWord);	
		}

		Word* pLastWord = dqWords.back ();
		dqWords.pop_back ();

		cout << endl;
		cout << dqWords << *pLastWord << endl;
		pair <double, double> pairTag = model.lm_Tag.Get (dqWords, pLastWord, Word::GetTagText, false);
		pair <double, double> pairLM = model.lm_Lexical.Get (dqWords, pLastWord, Word::GetWordText, true);
		pair <double, double> pairLMFull = model.lm_Lexical.GetFullScore (dqWords, pLastWord, Word::GetWordText, true);

		cout << "   LM      : " << pairLM.first << " : " << pairLM.second << endl;
		cout << "   LM full : " << pairLMFull.first << " : " << pairLMFull.second << endl;
		cout << "   Tag     : " << pairTag.first << " : " << pairTag.second << endl;
		cout << endl;
	}
}

