#include "Perceptron.h"
#include "FeatureEnumerator.h"
#include "Config.h"


//-------------------------------------------------------------------------
Perceptron::Perceptron (void)
{
	memset (arr_Weights, 0, FEATURE_COUNT * sizeof (double));
	for (int i = 0; i < FEATURE_COUNT; ++ i)
		arr_AllowUpdate [i] = 1;
	b_ReadOnly = false;
	b_Hierarchical = false;
}


Perceptron::~Perceptron (void) 
{
}
  

//-------------------------------------------------------------------------
void Perceptron::Update (FeatureVector& _rfvYGood, Feature_Vector_list_t& _lstBad)
{
	double arrTotals [FEATURE_COUNT];
	memset (arrTotals, 0, FEATURE_COUNT * sizeof (double));

	ITERATE_COLLECTION (Feature_Vector_list_t, _lstBad, iteBadVector) 
	{
		FeatureVector* pfvBad = *iteBadVector;
		for (int i = 0; i < FEATURE_COUNT; ++ i)
			arrTotals [i] += pfvBad->arr_Features [i];
	}

	double dBadFeatureVectorSize = (double)_lstBad.size ();
	for (int i = 0; i < FEATURE_COUNT; ++ i)
	{
		arr_Weights [i] += arr_AllowUpdate [i] 
						   * (_rfvYGood.arr_Features [i] - (arrTotals [i] / dBadFeatureVectorSize));
	}
}


//-------------------------------------------------------------------------
double Perceptron::Evaluate (FeatureVector& _rFeatures)
{
	double dScore = 0;

	for (int i = 0; i < FEATURE_COUNT; ++ i)
		dScore += arr_Weights [i] * _rFeatures.arr_Features [i];

	return dScore;
}


//-------------------------------------------------------------------------
void Perceptron::Save (const char* _zFileName, const char* _zTag)
{
	assert (false == b_ReadOnly);
    ofstream file (_zFileName, ios::app | ios::ate);

	file << _zTag 
		 << "   " << Config::Get ("TYPE")
		 << ".b" << Config::Get ("BEAM_SIZE") 
		 << "-d" << Config::Get ("DEPTH")
		 << "-i" << Config::Get ("ITERATIONS")
		 << "-l" << Config::Get ("TITLE_LENGTH")
		 << endl;

	file.precision (15);
	if (true == b_Hierarchical)
	{
		for (int i = 0; i < FEATURE_COUNT; ++ i)
		{
			String sName = FeatureEnumerator::TocIndexToName (i);
			if (sName != "")
				file << sName << ':' << arr_Weights [i] << ' ';
		}
	}
	else
	{
		for (int i = 0; i < FEATURE_COUNT; ++ i)
		{
			String sName = FeatureEnumerator::IndexToName (i);
			if (sName != "")
				file << sName << ':' << arr_Weights [i] << ' ';
		}
	}

	file << endl;

	file.close ();
}


//-------------------------------------------------------------------------
void Perceptron::LoadReadOnly (const char* _zFileName)
{
	b_ReadOnly = true;
	if( false == Load (_zFileName) )
	    throw;
}


//-------------------------------------------------------------------------
bool Perceptron::Load (const char* _zFileName)
{
    ifstream file (_zFileName); 
	if (false == file.is_open ())
	{
		cout << "Failed to open perceptron weights file '" << _zFileName << "'." << endl;
		return false;
	}
	else
	{
		char* pFileNameWithoutPath = (char*)(_zFileName + strlen (_zFileName));
		int iSlashCount = 0;
		while (_zFileName != pFileNameWithoutPath)
		{
			if ('/' == *pFileNameWithoutPath)
				++ iSlashCount;
			if (2 == iSlashCount)
				break;
			-- pFileNameWithoutPath;
		}
		cout << "  - reading weights from file '.." << pFileNameWithoutPath << "'" << endl;
	}

	bool bUseAverageWeights = false;
	if (true == b_Hierarchical)
		bUseAverageWeights = (1 == (int) Config::Get ("TOC_USE_AVERAGE_WEIGHTS"));
	else
		bUseAverageWeights = (1 == (int) Config::Get ("USE_AVERAGE_WEIGHTS"));

	if (true == bUseAverageWeights)
	{
		cerr << "  - using average weights." << endl;
		LoadAverageWeights (file);
	}
	else
	{
		cerr << "  - using last weights." << endl;
		LoadLastWeights (file);
	}

	// bool bReadOnly = b_ReadOnly;
	// b_ReadOnly = false;
	// String sFileName (_zFileName);
	// sFileName << ".dbg";
	// Save (sFileName, "DEBUG");
	// b_ReadOnly = bReadOnly;

	return true;
}


//-------------------------------------------------------------------------
void Perceptron::LoadLastWeights (ifstream& file)
{
    String sLine;
	String sFinalWeights;
    while (false == file.eof ()) 
	{
		getline (file, sLine);
		if (sLine != "")
			sFinalWeights = sLine;
	}
    file.close();

    Unigram_list_t lstFeatureAndWeights;
    sFinalWeights.Split (lstFeatureAndWeights);
    ITERATE_COLLECTION (Unigram_list_t, lstFeatureAndWeights, iteFeatureAndWeight) 
	{
		Unigram_list_t lstFeatWeightPair;
		iteFeatureAndWeight->Split (lstFeatWeightPair, ':');
		String sFeature = lstFeatWeightPair.front ();
		double dWeight = lstFeatWeightPair.back ();

		bool bDisallowUpdate = false;
		if (sFeature.at (0) == '+')
		{
			bDisallowUpdate = true;
			sFeature = sFeature.substr (1);
		}
		
		int iIndex = -1;
		if (true == b_Hierarchical)
			iIndex = FeatureEnumerator::TocNameToIndex (sFeature);
		else
			iIndex = FeatureEnumerator::NameToIndex (sFeature);
		if (-1 != iIndex)
		{
			arr_Weights [iIndex] = dWeight;
			if (true == bDisallowUpdate)
				arr_AllowUpdate [iIndex] = 0;
		}
		else if (dWeight != 0)
			cerr << "WARNING : Unknown feature weight with non-zero value : " 
				 << sFeature << " = " << dWeight << endl;
    }
}


//-------------------------------------------------------------------------
void Perceptron::LoadAverageWeights (ifstream& file)
{
	long lLinesToAverage = Config::Get ("LINES_TO_AVERAGE");
	long lLineToStopAt = Config::Get ("ITERATION_TO_STOP_AT");

	// laodup all the weight vectors from the file...		
    String sLine;
	long lLineCount = 0;
	list <String> lstLines;
    while (false == file.eof ()) 
	{
		getline (file, sLine);
		if (sLine == "")
			continue;
		if (sLine.at (0) == '[')
			continue;

		lstLines.push_back (sLine);
		if (0 != lLinesToAverage)
		{
			if ((long) lstLines.size () > lLinesToAverage)
				lstLines.pop_front ();
		}

		++ lLineCount;
		if ((0 != lLineToStopAt) && (lLineCount >=  lLineToStopAt))
		{
			cout << "  - stopping averaging at iteration " << lLineCount << "." << endl;
			break;
		}
	}
    file.close();

	FeatureValue_map_t	mapValues;
	FeatureValue_map_t	mapCounts;

	lLineCount = 0;
	ITERATE_COLLECTION (list<String>, lstLines, iteLine)
	{
		String sLine = *iteLine;

		Unigram_list_t lstFeature;
		sLine.Split (lstFeature);
		ITERATE_COLLECTION (Unigram_list_t, lstFeature, ite) 
		{
			Unigram_list_t lstValue;
			ite->Split (lstValue, ':');
			String& rName = lstValue.front ();
			double dWeight = lstValue.back ();

			// accumulate weights...						
			FeatureValue_map_t::iterator	iteValue;
			iteValue = mapValues.find (rName);
			if (iteValue == mapValues.end ())
				mapValues.insert (make_pair (rName, dWeight));
			else
				iteValue->second += dWeight;

			// increment counts...							
			FeatureValue_map_t::iterator	iteCount;
			iteCount = mapCounts.find (rName);
			if (iteCount == mapCounts.end ())
				mapCounts.insert (make_pair (rName, 1));
			else
				++ iteCount->second;
		}
		++ lLineCount;
	}


	// create the average perceptron weights...				
	ITERATE_COLLECTION (FeatureValue_map_t, mapCounts, iteCount)
	{
		FeatureValue_map_t::iterator	iteValue;
		iteValue = mapValues.find (iteCount->first);

		double dAverageValue = 0;
		if (iteValue != mapValues.end ())
			dAverageValue = iteValue->second / iteCount->second;

		String sFeature = iteCount->first;
		bool bDisallowUpdate = false;
		if (sFeature.at (0) == '+')
		{
			bDisallowUpdate = true;
			sFeature = sFeature.substr (1);
		}

		int iIndex = -1;
		if (true == b_Hierarchical)
			iIndex = FeatureEnumerator::TocNameToIndex (sFeature);
		else
			iIndex = FeatureEnumerator::NameToIndex (sFeature);
		if (-1 != iIndex)
		{
			arr_Weights [iIndex] = dAverageValue;
			if (true == bDisallowUpdate)
				arr_AllowUpdate [iIndex] = 0;
		}
		else if (dAverageValue != 0)
			cerr << "WARNING : Unknown feature weight with non-zero value : " 
				 << sFeature << " = " << dAverageValue << endl;
	}

	cout << "  - averages calculated over " << lLineCount << " generations of weight vectors." << endl;
}


